/*
* LDAP_TCL.cpp, v1.1 1/20/2003, Scott J. Nichols
*
* This is the C source code for a TCL 8.3 library extension (DLL)
* This TCL extension performs asynchronous LDAP searches
*
* Revisions:
* v1.2 3/13/2003 - Scott Nichols
* Added ldapssl_client_init and set LDAP version three in the function
* Ldapssl_Init_ObjCmd.
*
* v1.3 3-21-2003 - Scott Nichols
* Added ldapssl_install_routines if the objc passed objects is equal to three.
*
*
* v1.4 4-1-2003 - Scott Nichols
* Added ldap_search_ext_s for doing synchronous searches and removed ldapssl_install_routines
*
*/

#define USE_NON_CONST
#define TCL_USE_STUBS

#include "ldap_Class.h"

/*
 * Creates a new Tcl command, returns an LDAP *
 * handle to the TCL Intepreter.
 */
static int	Ldap_Init_ObjCmd _ANSI_ARGS_((ClientData clientData,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]))
{
	LDAP	*ld;
	char	sHandle[24];
	char	*host;
	int		port, error;

	static int iHandles = 0;

	/* Parse the Server name */
	host = Tcl_GetStringFromObj(objv[1], (int *) NULL);

	/* Parse the LDAP port */
	error = Tcl_GetIntFromObj(interp, objv[2], &port);
	if (error != TCL_OK) return error;

	/* Creates a new LDAP connection */
	if ( (ld = ldap_init( host, port )) == NULL ) {
		Tcl_AppendResult(interp, ldap_err2string(ldap_get_lderrno((ld),NULL,NULL)),
			Tcl_PosixError(interp), (char *) NULL);
		return TCL_ERROR;
	}

	/* Used for Dynamic Handles*/
	iHandles++;

	/* String representation of LDAP Handle*/
	sprintf(sHandle, "LDAP%d", iHandles);

	/* Creates a new dynamic command each time ldap_init is called from a TCL script*/
	Tcl_CreateObjCommand( interp,sHandle, Ldap_Handle_ObjCmd,
		(ClientData)ld, (Tcl_CmdDeleteProc *) NULL);

	/* Return the LDAP handle to calling TCL script*/
	Tcl_SetResult( interp, sHandle, TCL_VOLATILE);

	return TCL_OK;
}

/*
*  This function is used to intialize LDAP for SSL.
*  NOTE: Currently, only weak SSL is supported.
*  Please review the LDAP_ssl.h header file for more info.
*/
static int	Ldapssl_Init_ObjCmd _ANSI_ARGS_((ClientData clientData,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]))
{
	LDAP	*ld;
	char	sHandle[24];
	char	*host;
	char    *certificatePath;
	int		port, error,iAuthMode=1;
	static int iHandles = 0;

	/* Did the TCL programmer pass all the needed arguments */
	if (objc < 5)
	{
		Tcl_WrongNumArgs(interp,1,objv,
					   "Host Port CertificatePath Authmode");
		return TCL_ERROR;
	}

	/* Parse the Server name */
	host = Tcl_GetStringFromObj(objv[1], (int *) NULL);

	/* Parse the LDAP port */
	error = Tcl_GetIntFromObj(interp, objv[2], &port);
	if (error != TCL_OK) return error;


	/* Try the certificate database */
	/* Parse the ceritificate path */
	certificatePath = Tcl_GetStringFromObj(objv[3], (int *) NULL);

	// Initialize the client.
	if (ldapssl_client_init(certificatePath,NULL) < 0)
	{
		Tcl_AppendResult(interp, "ldapssl_client_init error",
			Tcl_PosixError(interp), (char *) NULL);
		return TCL_ERROR;

	}


	/* Valid authentification mode options */
	static char *options[] = {
		"LDAPSSL_AUTH_WEAK",
		"0",
		"LDAPSSL_AUTH_CERT",
		"1",
		"LDAPSSL_AUTH_CNCHECK",
		"2",
		NULL
	};

	enum options {
		LDAPSSLAUTHWEAK,
		ZERO,
		LDAPSSLAUTHCERT,
		ONE,
		LDAPSSLAUTHCNCHECK,
		TWO
	};

	int index =0;
	/* Parse the sub-command */
	if (Tcl_GetIndexFromObj(interp, objv[4],options, "option", 0,
		&index) != TCL_OK) {
		return TCL_ERROR;
	}

	/* Switch on the sub-command */
	switch ((enum options) index)
	{
		case LDAPSSLAUTHWEAK:
		case ZERO:
		{
			iAuthMode = LDAPSSL_AUTH_WEAK;
			break;
		}
		case LDAPSSLAUTHCERT:
		case ONE:
		{
			iAuthMode = LDAPSSL_AUTH_CERT;
			break;
		}
		case LDAPSSLAUTHCNCHECK:
		case TWO:
		{
			iAuthMode = LDAPSSL_AUTH_CNCHECK;
			break;
		}
	}

	/* Creates a new LDAP connection */
	if ( (ld = ldapssl_init( host, port,iAuthMode )) == NULL ) {
		Tcl_AppendResult(interp, ldap_err2string(ldap_get_lderrno((ld),NULL,NULL)),
			Tcl_PosixError(interp), (char *) NULL);
		return TCL_ERROR;
	}

	/* Use LDAP version 3 */
	int iVersion = LDAP_VERSION3;
	if (ldap_set_option(ld,LDAP_OPT_PROTOCOL_VERSION, &iVersion) < 0)
	{
		Tcl_AppendResult(interp, ldap_err2string(ldap_get_lderrno((ld),NULL,NULL)),
			Tcl_PosixError(interp), (char *) NULL);
		return TCL_ERROR;

	}

	/* Used for Dynamic Handles*/
	iHandles++;

	/* String representation of LDAP Handle*/
	sprintf(sHandle, "LDAP_SSL_%d", iHandles);

	/* Creates a new dynamic command each time ldap_init is called from a TCL script*/
	Tcl_CreateObjCommand( interp,sHandle, Ldap_Handle_ObjCmd,
		(ClientData)ld, (Tcl_CmdDeleteProc *) NULL);

	/* Return the LDAP handle to calling TCL script*/
	Tcl_SetResult( interp, sHandle, TCL_VOLATILE);

	return TCL_OK;
}

/* Dynamic LDAP Handle Command*/
static int	Ldap_Handle_ObjCmd _ANSI_ARGS_((ClientData data,
		    Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[]))
{
	LDAP	*ld = (LDAP *)data;
	int index;
	int	msgid;
	int lArgc;
	char **attrs;
	LDAPMessage	*result, *e;
	char			 **vals;
	Tcl_Obj		*obj_result;
	Tcl_Obj		   *obj_row;
	Tcl_Obj 	   *obj_col;
	BerElement		   *ber;
	char			*a, *dn;
	char sMsgID[256];
	int rc,i,bAll=1;

	// Constuct the LDAP object
	LDAP_CLASS objLDAP(ld);

	if (objc < 2)
	{
		Tcl_WrongNumArgs(interp,2,objv,
					   "Usage : $ldap_handle ldap_simple_bind_s or ldap_search or ldap_unbind or ldap_result or ldap_init");
		return TCL_ERROR;
	}

	/* Valid LDAP sub command options */
    static char *options[] = {
			"ldap_simple_bind_s",
			"ldap_search",
			"ldap_search_ext_s",
			"ldap_unbind",
			"ldap_result",
			"ldap_unbind_s",
			"ldap_set_option",
			NULL
	};

    enum options {
		LDAP_BIND_S,
		LDAP_SEARCH,
		LDAP_SEARCH_EXT_S,
		LDAP_UNBIND,
		LDAP_RESULT,
		LDAP_UNBIND_S,
		LDAP_SET_OPTION
    };

	/* Parse the LDAP sub-command */
    if (Tcl_GetIndexFromObj(interp, objv[1], options, "option", 0,
	    &index) != TCL_OK) {
		return TCL_ERROR;
    }

    /* Switch on the sub-command */
	switch ((enum options) index)
    {
		case LDAP_BIND_S:
		{
			char *ldapDN = NULL;
			char *ldapPassword = NULL;

			if (objc > 2)
				ldapDN = Tcl_GetStringFromObj(objv[2], (int *) NULL);

			if (objc > 3)
				ldapPassword = Tcl_GetStringFromObj(objv[3], (int *) NULL);

			if(ld !=NULL)
			{
				if (  objLDAP.LDAP_BIND(ldapDN,ldapPassword) != LDAP_SUCCESS )
				{
					ldap_perror (ld, "ldap_simple_bind_s");
					interp->result = ldap_err2string(ldap_get_lderrno((ld),NULL,NULL));
					return TCL_ERROR;
				}
				else
				{
					interp->result = "SUCCEEDED";
				}
			}
			else
			{
				interp->result = "ERROR {LDAP handle invalid}";
				return TCL_ERROR;
			}
			break;
		}
		case LDAP_SEARCH:
		{

			if (objc < 4)
			{
				Tcl_WrongNumArgs(interp,
								 2,
								 objv,
							     "Usage : $ldap_handle ldap_search base searchfilter [TCL List: Arguments]"
							     );
				return TCL_ERROR;
			}

			/* Parse the TCL list attributes, if it exists */
			if (objc > 4) {
				if((Tcl_SplitList(interp, Tcl_GetStringFromObj(objv[4], (int *) NULL), &lArgc, &attrs)) != TCL_OK)
				{
					Tcl_AppendResult(interp, "ERROR: TCL LIST argument incorrect!", (char *) NULL);
					return TCL_ERROR;
				}
			} else {
			  attrs = NULL;
			}

			/*Search for all entries*/
			if ((msgid=objLDAP.LDAP_SEARCH(Tcl_GetStringFromObj(objv[2],
			                       (int *) NULL),LDAP_SCOPE_SUBTREE,
			                       Tcl_GetStringFromObj(objv[3],
			                       (int *) NULL)
			                       ,attrs)
			                       ) == -1)
			{
				ldap_perror(ld, "ldap_search");
				interp->result = ldap_err2string(ldap_get_lderrno((ld),NULL,NULL));
				return TCL_ERROR;
			}

			sprintf(sMsgID,"%i",msgid);
			interp->result = sMsgID;

			break;
		}
		case LDAP_SEARCH_EXT_S:
		{

			if (objc < 4)
			{
				Tcl_WrongNumArgs(interp,
								 2,
								 objv,
							     "Usage : $ldap_handle ldap_search base searchfilter [TCL List: Arguments]"
							     );
				return TCL_ERROR;
			}

			/* Parse the TCL list attributes, if it exists */
			if (objc > 4) {
				if((Tcl_SplitList(interp, Tcl_GetStringFromObj(objv[4], (int *) NULL), &lArgc, &attrs)) != TCL_OK)
				{
					Tcl_AppendResult(interp, "ERROR: TCL LIST argument incorrect!", (char *) NULL);
					return TCL_ERROR;
				}
			} else {
			  attrs = NULL;
			}

			timeval tv;
			tv.tv_sec = 300;
			tv.tv_usec = NULL;

			/* search all entries */
			if ( ldap_search_ext_s( ld,
							    Tcl_GetStringFromObj(objv[2],(int *) NULL),
							    LDAP_SCOPE_SUBTREE,
								Tcl_GetStringFromObj(objv[3],(int *) NULL),
								attrs,
								0,
								NULL,
								NULL,
								&tv,
								0,
								&result
							  ) != LDAP_SUCCESS )
			{
				ldap_perror(ld, "ldap_search_ext_s");
				interp->result = ldap_err2string(ldap_get_lderrno((ld),NULL,NULL));
				return TCL_ERROR;
			}

			// Set the first TCL field to the return code
			obj_result = Tcl_GetObjResult(interp);
			for ( e=ldap_first_entry(ld,result); e !=NULL;e = ldap_next_entry(ld,e))
			{
				obj_row = Tcl_NewListObj(0,NULL);
				if ((dn=ldap_get_dn(ld,e)) !=NULL)
				{
					ldap_memfree(dn);
				}

				for ( a=ldap_first_attribute(ld,e,&ber);a !=NULL; a=ldap_next_attribute(ld,e,ber))
				{
					if ((vals=ldap_get_values(ld,e,a)) != NULL)
					{
						for (i=0;vals[i] !=NULL; i++)
						{

							obj_col = Tcl_NewStringObj(a,strlen(a));
							Tcl_ListObjAppendElement(interp,obj_row,obj_col);

							obj_col = Tcl_NewStringObj(vals[i],strlen(vals[i]));
							Tcl_ListObjAppendElement(interp,obj_row,obj_col);

						}
						ldap_value_free(vals);
					}
						ldap_memfree(a);
				}
				if (ber !=NULL)
				{
					ber_free(ber,0);
				}
				Tcl_ListObjAppendElement(interp,obj_result,obj_row);
			}

			break;
		}
 		case LDAP_UNBIND:
		{
			if ( objLDAP.LDAP_UNBIND() == LDAP_SUCCESS)
			{
				Tcl_DeleteCommand(interp,Tcl_GetStringFromObj(objv[0], (int *) NULL));
				interp->result = "SUCCEEDED";
			}
			else
				interp->result = "FAILED";
			break;
		}
 		case LDAP_UNBIND_S:
		{
			if ( ldap_unbind_s( ld ) == LDAP_SUCCESS)
			{
				Tcl_DeleteCommand(interp,Tcl_GetStringFromObj(objv[0], (int *) NULL));
				interp->result = "SUCCEEDED";
			}
			else
				interp->result = "FAILED";
			break;
		}


		case LDAP_RESULT:
		{
			timeval tv;
			int seconds;
			tv.tv_sec = NULL;
			tv.tv_usec = NULL;

			/* Verify the user has sent all the needed arguments. */
			if (objc < 4)
			{
				interp->result = "FAILED {Usage : $LDAP_HANDLE ldap_result \"Message ID\" \"timeout\" [bAll]}";
				return TCL_ERROR;
			}

			/* Parse the LDAP Search Message ID */
			if (Tcl_GetIntFromObj(interp, objv[2], &msgid) != TCL_OK)
				return TCL_ERROR;

			/* Parse the timeout */
			if (Tcl_GetIntFromObj(interp, objv[3], &seconds) != TCL_OK)
					return TCL_ERROR;

			// Set the timeout structure.
			tv.tv_sec = seconds;
			tv.tv_usec = 0;

			/* Parse the LDAP bAll Parameter */
			if (objc > 4)
			{
				if (Tcl_GetIntFromObj(interp, objv[4], &bAll) != TCL_OK)
					return TCL_ERROR;
			}

			// Call the objFunction
			rc = objLDAP.LDAP_RESULT(msgid,&bAll,&tv,&result);

			// Was there an error or timeout?
			if (rc == -1 || rc == 0)
			{
				ldap_perror(ld,"ldap_result");
				interp->result = ldap_err2string(ldap_get_lderrno((ld),NULL,NULL));
				return TCL_ERROR;
			}
			else if (rc == LDAP_RES_SEARCH_RESULT || rc== LDAP_RES_SEARCH_ENTRY)
			{
				// Set the first TCL field to the return code
				obj_result = Tcl_GetObjResult(interp);
				obj_row = Tcl_NewListObj(0,NULL);
				obj_col = Tcl_NewIntObj(rc);
				Tcl_ListObjAppendElement(interp,obj_row,obj_col);
				Tcl_ListObjAppendElement(interp,obj_result,obj_row);

				// Check for search errors
				rc = ldap_result2error(ld,result,0);

				if( rc == LDAP_SUCCESS)
				{
					for ( e=ldap_first_entry(ld,result); e !=NULL;e = ldap_next_entry(ld,e))
					{
						obj_row = Tcl_NewListObj(0,NULL);
						if ((dn=ldap_get_dn(ld,e)) !=NULL)
						{
							ldap_memfree(dn);
						}

						for ( a=ldap_first_attribute(ld,e,&ber);a !=NULL; a=ldap_next_attribute(ld,e,ber))
						{
							if ((vals=ldap_get_values(ld,e,a)) != NULL)
							{
								for (i=0;vals[i] !=NULL; i++)
								{

									obj_col = Tcl_NewStringObj(a,strlen(a));
									Tcl_ListObjAppendElement(interp,obj_row,obj_col);

									obj_col = Tcl_NewStringObj(vals[i],strlen(vals[i]));
									Tcl_ListObjAppendElement(interp,obj_row,obj_col);

								}
								ldap_value_free(vals);
							}
								ldap_memfree(a);
						}
						if (ber !=NULL)
						{
							ber_free(ber,0);
						}
						Tcl_ListObjAppendElement(interp,obj_result,obj_row);
					}
				}
				else
				{
					ldap_perror(ld,"ldap_result");
					interp->result = ldap_err2string(rc);
					return TCL_ERROR;
				}
			}
			else
			{
				sprintf(sMsgID,"%i",rc);
				interp->result = sMsgID;
			}

			break;
		}
		
		case LDAP_SET_OPTION:
		{
		
			if (objc < 4)
			{
				Tcl_WrongNumArgs(interp,
								 2,
								 objv,
							     "Usage : $ldap_handle ldap_set_option [option] [value]"
							     );
				return TCL_ERROR;
			}						
			
			int lopt = 0;
			
			if (strcmp("LDAP_OPT_API_INFO",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_API_INFO;
			else if (strcmp("LDAP_OPT_DESC",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_DESC;
			else if (strcmp("LDAP_OPT_DEREF",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_DEREF;
			else if (strcmp("LDAP_OPT_SIZELIMIT",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_SIZELIMIT;
			else if (strcmp("LDAP_OPT_TIMELIMIT",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_TIMELIMIT;
			else if (strcmp("LDAP_OPT_REFERRALS",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_REFERRALS;			
			else if (strcmp("LDAP_OPT_RESTART",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_RESTART;
			else if (strcmp("LDAP_OPT_PROTOCOL_VERSION",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_PROTOCOL_VERSION;
			else if (strcmp("LDAP_OPT_SERVER_CONTROLS",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_SERVER_CONTROLS;
			else if (strcmp("LDAP_OPT_CLIENT_CONTROLS",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_CLIENT_CONTROLS;
			else if (strcmp("LDAP_OPT_API_FEATURE_INFO",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_API_FEATURE_INFO;
			else if (strcmp("LDAP_OPT_HOST_NAME",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_HOST_NAME;
			else if (strcmp("LDAP_OPT_ERROR_NUMBER",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_ERROR_NUMBER;
			else if (strcmp("LDAP_OPT_ERROR_STRING",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_ERROR_STRING;
			else if (strcmp("LDAP_OPT_MATCHED_DN",Tcl_GetStringFromObj(objv[2], (int *) NULL)) == 0)
				lopt = LDAP_OPT_MATCHED_DN;
			else
			{
				Tcl_WrongNumArgs(interp,
								 2,
								 objv,
							     "Unknown LDAP option"
							     );
				return TCL_ERROR;
			}
			int lvalue = 0;
			if (Tcl_GetIntFromObj(interp, objv[3], &lvalue) != TCL_OK)
				return TCL_ERROR;
			
			
			ldap_set_option(objLDAP.ld, lopt, &lvalue);

			
			return TCL_OK;

		}
		default:
		{
			interp->result = "ERROR: Unknown sub-command";
			return TCL_ERROR;
		}
	}
	return TCL_OK;

}

/* Main Routine in the TCL Extension DLL*/
int Ldap_tcl_Init(Tcl_Interp *interp)
{
	#if TCL_MAJOR_VERSION == 8 && TCL_MINOR_VERSION >= 1
		// Does the TCL interpreter support version 8.3 of TCL?
		if (Tcl_InitStubs(interp,"8.3",0) == NULL)
			return TCL_ERROR;
	#endif


	Tcl_CreateObjCommand(interp, "ldap_init", Ldap_Init_ObjCmd,
		      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	Tcl_CreateObjCommand(interp, "ldapssl_init", Ldapssl_Init_ObjCmd,
		      (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

	return (Tcl_PkgProvide(interp,"LDAP","1.1") == TCL_ERROR ? TCL_ERROR : TCL_OK);
}